/*************************************************************
sock.C
HummingBird Project - Deborah A. Frincke, PhD.
Spring 1998
Coded by: Jamie Marconi
Description: This file is the implementation fo the base socket
class this is inherited by sock_serv and sock_cli.
***************************************************************/
/*
 Copyright (C) 1997-1998 The HummingBird Project - hummer@cs.uidaho.edu

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "sock.h"

/********
Function: sock::sock
Purpose:  Basic constructor
********/

sock::sock()
{
  sock = -1;
  kerberized = -1;
  xmit_time = 300;

  sin_len = sizeof (struct sockaddr_in);
  rmt_sockaddr = new (struct sockaddr_in);
  loc_sockaddr = new (struct sockaddr_in);

  service = NULL;

#ifdef KRB5
  retval = krb5_init_context(&context);
  if (retval) {
    com_err("sock::sock", retval, "while initializing krb5");
  }

  auth_context = NULL;
#endif
 
}

/*******
Function: sock::~sock
Purpose:  Basic destructor
*******/

sock::~sock()
{
  sock = -1;

  delete rmt_sockaddr;
  delete loc_sockaddr;

#ifdef KRB5
  if (context)
    krb5_free_context(context);
  context = NULL;
#endif
}


/*******
Function: sock::create()
Purpose:  Call the library initializaton functions. Also creates a 
socket file descriptor.
*******/

int sock::create()
{

#ifdef KRB5
  // Initialize the context for the object
  retval = krb5_init_context(&context);
  if (retval) {
    com_err("sock::create", retval, "while initializing krb5");
    return 0;
  }

  // Check to see that CRC32 is supported
  if (!valid_cksumtype(CKSUMTYPE_CRC32)) {
    com_err("sock::create()", KRB5_PROG_SUMTYPE_NOSUPP, "while using CRC-32");
    return 0;
  }
#endif

  signal(SIGPIPE, SIG_IGN);

  // Create the socket
  if ( (sock = socket(AF_INET, SOCK_STREAM, 0) ) == -1 ) {
#ifdef VERBOSE
    Print_Error("sock::create", "%s: while creating socket", strerror(errno));
#endif
    return 0;
  }
  
  return 1;
}

/*******
Function: sock::close
Purpose:  Frees the files descriptor and closes the socket connection.
*******/ 
int sock::close()
{
#ifdef DEBUG_SOCK
  Print_Error("sock::close","Closing file descriptor %d", sock);
#endif
  if (::close(sock) == -1) 
    return 0;

  return 1;
}

/******
Function: sock::get_fd
Purpose:  Returns the file descriptor that is associated with this object.
******/

int sock::get_fd()
{
  return sock;
}

/******
Function: sock::set_fd
Purpose:  Set the file descriptor associated with this object.
*******/

void sock::set_fd(int new_fd)
{
  sock = new_fd;
}

/**********
Function: sock::is_kerb5
Purpose:  Determine if communication will be done with kerberized sockets
***********/

int sock::is_kerb5()
{
  int buflen = strlen(KERBERIZED_SOCKET);
  char *buf = new char[buflen];
  

  bzero((void *) buf, buflen);
	 
#ifdef KRB5
  kerberized = 1;
#else
  kerberized = 0;
#endif
  
  if (kerberized) {
    ::write(sock, KERBERIZED_SOCKET, buflen);
  }
  else {
    ::write(sock, UN_KERBERIZED_SOCKET, buflen);
  }

  if (read_wait()) {
    ::read(sock, buf, buflen);
  }
  
  if (buf[0] == 'K') {
    kerberized = kerberized & 1;
  }
  else {
    kerberized = 0;
  }

  delete [] buf;

  return kerberized;
}

/**********
Function: sock::send_auth
Purpose:  Send authentication material to server and authenicate remote host with pservice
          returns 0 on error, 1 on success
***********/

int sock::send_auth(char *pservice)
{
  service = strdup(pservice);
  
  return send_auth();
}

/**********
Function: sock::send_auth
Purpose:  Send authenication material to server and authenicate remote host.
          returns 0 on error, 1 on success
**********/ 

int sock::send_auth()
{
  if (getpeername(sock, (struct sockaddr *) rmt_sockaddr, &sin_len) < 0) {
#ifdef VERBOSE
    Print_Error("sock::send_auth","getpeername error");
#endif
    return NULL;
  }
  
#ifdef DEBUG_SOCK
  Print_Error("sock::send_auth", "rmt_sockaddr.sin_addr is %s", inet_ntoa(rmt_sockaddr->sin_addr));
  Print_Error("sock::send_auth", "rmt_sockaddr.sin_port is %d", ntohs(rmt_sockaddr->sin_port)); 
#endif
  
  // Fill out a sockaddr_in for the local hostname
  bzero((char *)loc_sockaddr, sin_len);
  loc_sockaddr->sin_family = AF_INET;
#ifdef BROKEN_STREAMS_SOCKETS
  if (gethostname(full_loc_hname, sizeof(full_loc_hname)) < 0) {
#ifdef VERBOSE
    Print_Error("sock::send_auth","%s while getting local host name", strerror(errno));
#endif
    return NULL;
  }
  
  struct hostent *host;
  if ((host = gethostbyname(full_loc_hname)) == (struct hostent *)0) {
#ifdef VERBOSE
    Print_Error("sock::send_auth", "%s: unknown host", full_loc_hname);
#endif
    return NULL;
  }
  memcpy((char *) &loc_sockaddr->sin_addr, host->h_addr, sizeof(loc_sockaddr->sin_addr));
#endif /* BROKEN_STREAMS_SOCKETS */ 
  
  int sockaddr_in_len = sizeof(struct sockaddr_in);
  if (getsockname(sock, (struct sockaddr *) loc_sockaddr, &sockaddr_in_len) < 0) {
#ifdef VERBOSE
    Print_Error("sock::send_auth", "%s while getting sock name", strerror(errno));
#endif
    ::close(sock);
    return NULL;
  }
  
#ifdef DEBUG_SOCK
  Print_Error("sock::send_auth", "loc_sockaddr.sin_addr is %s", inet_ntoa(loc_sockaddr->sin_addr));
  Print_Error("sock::send_auth", "loc_sockaddr.sin_port is %d", ntohs(loc_sockaddr->sin_port));
#endif

  if (is_kerb5()) {
#ifdef KRB5
    krb5_ccache ccache = NULL;
    krb5_principal client;
    krb5_principal server;
    
    if (retval = krb5_cc_default(context, &ccache)) {
      com_err("sock::send_auth", retval, "while getting default ccache");
      return 0;
    }
    
    if (!keytab_name && !principal_name) {
      if (retval = krb5_cc_get_principal(context, ccache, &client)) {
	com_err("sock::send_auth", retval, "while getting client principal name");
	return 0;
      }
    }
    else {
      
      if (retval = krb5_cc_get_principal(context, ccache, &client)) {
	
	krb5_data tgtname = {
	  0,
	  KRB5_TGS_NAME_SIZE,
	  KRB5_TGS_NAME
	};
	
	krb5_preauthtype * preauth = NULL;
	int options = KRB5_DEFAULT_OPTIONS;
	int option;
	int errflg = 0;
	krb5_creds my_creds;
	krb5_address *null_addr = (krb5_address *)0;
	krb5_address **addrs = (krb5_address **)0;
	krb5_keytab keytab = NULL;
	char  *client_name;
	
	retval = krb5_kt_resolve(context, keytab_name, &keytab);
	if (retval != 0) {
	  com_err("sock::send_auth", retval, "resolving keytab %s",
		  keytab_name);
	  errflg++;
	}
	
	if ((retval = krb5_parse_name (context, principal_name, &client))) {
	  com_err ("sock::send_auth", retval, "when parsing name %s",principal_name);
	  return 0;
	}
	
	if ((retval = krb5_unparse_name(context, client, &client_name))) {
	  com_err ("sock::send_auth", retval, "when unparsing name");
	  return 0;
	}
	
	bzero((char *)&my_creds, sizeof(my_creds));
	
	my_creds.client = client;
	
	if((retval = krb5_build_principal_ext(context, &server,
					      krb5_princ_realm(context, client)->length,
					      krb5_princ_realm(context, client)->data,
					      tgtname.length, tgtname.data,
					      krb5_princ_realm(context, client)->length,
					      krb5_princ_realm(context, client)->data,
					      0))) {
	  com_err("sock::send_auth", retval, "while building server name");
	  return 0;
	}
	
	my_creds.server = server;
	
	retval = krb5_get_in_tkt_with_keytab(context, options, addrs,
					     NULL, preauth, keytab, 0,
					     &my_creds, 0);
	
	if (retval) {
	  if (retval == KRB5KRB_AP_ERR_BAD_INTEGRITY)
#ifdef VERBOSE
	    Print_Error("sock::send_auth", "Password incorrect");
#endif
	  else
	    com_err ("sock::send_auth", retval, "while getting initial credentials");
	  return 0;
	}
	
	retval = krb5_cc_initialize (context, ccache, client);
	if (retval != 0) {
	  com_err ("sock::send_auth", retval, "when initializing cache ");
	  return 0;
	}
	
	retval = krb5_cc_store_cred(context, ccache, &my_creds);
	if (retval) {
	  com_err ("sock::send_auth", retval, "while storing credentials");
	  return 0;
	}
      }
      else {
#ifdef DEBUG_SOCK
	Print_Error("sock::send_auth", "Using Default Cache name");
#endif
      }
    }
    
    if (!service) {
#ifdef DEBUG_SOCK
      Print_Error("sock::send_auth","Proceeding with out service name");
#endif
    }
    
    if (retval = krb5_sname_to_principal(context, full_rmt_hname, service,
					 KRB5_NT_SRV_HST, &server)) {
      com_err("sock::send_auth", retval, "while creating server name for %s/%s",
	      full_rmt_hname, service);
      return 0;
    }
    
    krb5_data cksum_data;
    cksum_data.data = full_rmt_hname;
    cksum_data.length = strlen(full_rmt_hname);
    
    krb5_error *err_ret;
    krb5_ap_rep_enc_part *rep_ret;
    retval = krb5_sendauth(context, &auth_context, (krb5_pointer) &sock,
			   KRB5_SOCK_VERSION, client, server,
			   AP_OPTS_MUTUAL_REQUIRED,
			   &cksum_data,
			   0,		/* no creds, use ccache instead */
			   ccache, &err_ret, &rep_ret, NULL);
    
    krb5_free_principal(context, server);	/* finished using it */ 
    krb5_free_principal(context, client);
    
    if (retval && retval != KRB5_SENDAUTH_REJECTED) {
      com_err("sock::send_auth", retval, "while using sendauth");
      return 0;
    }
    if (retval == KRB5_SENDAUTH_REJECTED) {
      /* got an error */
#ifdef DEBUG_SOCK
      Print_Error("sock::send_auth","sendauth rejected, error reply is:\n\t\"%*s\"\n",
		  err_ret->text.length, err_ret->text.data);
#endif
      return 0;
    } else if (rep_ret) {
      /* got a reply */
#ifdef DEBUG_SOCK
      Print_Error("sock::send_auth", "sendauth succeeded");
#endif
    } else {
      com_err("sock::send_auth", 0, "no error or reply from sendauth!");
      return 0;
    }
    
    krb5_address loc_addr;
    krb5_address rmt_addr;
    
    loc_addr.addrtype = ADDRTYPE_IPPORT;
    loc_addr.length = sizeof(loc_sockaddr->sin_port);
    loc_addr.contents = (krb5_octet *) &loc_sockaddr->sin_port;
    
    rmt_addr.addrtype = ADDRTYPE_IPPORT;
    rmt_addr.length = sizeof(rmt_sockaddr->sin_port);
    rmt_addr.contents = (krb5_octet *) &rmt_sockaddr->sin_port;
    
    if ((retval = krb5_auth_con_setports(context, auth_context,
					 &loc_addr, &rmt_addr))) {
      com_err("sock::send_auth", retval, "while setting local port\n");
      return 0;
    }
    
    loc_addr.addrtype = ADDRTYPE_INET;
    loc_addr.length = sizeof(loc_sockaddr->sin_addr);
    loc_addr.contents = (krb5_octet *) &loc_sockaddr->sin_addr;
    
    rmt_addr.addrtype = ADDRTYPE_INET;
    rmt_addr.length = sizeof(rmt_sockaddr->sin_addr);
    rmt_addr.contents = (krb5_octet *) &rmt_sockaddr->sin_addr;
    
    if ((retval = krb5_auth_con_setaddrs(context, auth_context,
					 &loc_addr, &rmt_addr))) {
      com_err("sock::send_auth", retval, "while setting local addr\n");
      return 0;
    }
    
    krb5_address *portlocal_addr;
    if ((retval = krb5_gen_portaddr(context, &loc_addr,
				    (krb5_pointer) &loc_sockaddr->sin_port,
				    &portlocal_addr))) {
      com_err("sock::send_auth", retval, "while generating port address");
      return 0;
    }
    
    char *cp;
    if ((retval = krb5_gen_replay_name(context, portlocal_addr,
				       REPLAY_UNIQUE_NAME, &cp))) {
      com_err("sock::send_auth", retval, "while generating replay cache name");
      return 0;
    }
    
    krb5_data rcache_name;
    rcache_name.length = strlen(cp);
    rcache_name.data = cp;
    
    krb5_rcache rcache;
    if ((retval = krb5_get_server_rcache(context, &rcache_name, &rcache))) {
      com_err("sock::send_auth", retval, "while getting server rcache");
      return 0;
    }

    /* set auth_context rcache */
    krb5_auth_con_setrcache(context, auth_context, rcache);
    
    krb5_xfree(cp);
    
#endif /* KRB5 */
  }
  else {
#ifdef DEBUG_SOCK
    Print_Error("sock::send_auth", "Not using Kerberos");
#endif
  }
  return 1;
}

/********
Function: sock::recv_auth
Purpose: Recieve and process Authorization information
********/

char * sock::recv_auth(char *keytab_name)
{
  if (getpeername(sock, (struct sockaddr *) rmt_sockaddr, &sin_len) < 0) {
#ifdef VERBOSE
    Print_Error("sock::recv_auth","getpeername error");
#endif
    return NULL;
  }
  
#ifdef DEBUG_SOCK
  Print_Error("sock::recv_auth", "rmt_sockaddr.sin_addr is %s", inet_ntoa(rmt_sockaddr->sin_addr));
  Print_Error("sock::recv_auth", "rmt_sockaddr.sin_port is %d", ntohs(rmt_sockaddr->sin_port)); 
#endif
  
  // Fill out a sockaddr_in for the local hostname
  bzero((char *)loc_sockaddr, sin_len);
  loc_sockaddr->sin_family = AF_INET;
#ifdef BROKEN_STREAMS_SOCKETS
  if (gethostname(full_loc_hname, sizeof(full_loc_hname)) < 0) {
#ifdef VERBOSE
    Print_Error("sock::recv_auth","%s while getting local host name", strerror(errno));
#endif
    return NULL;
  }
  
  struct hostent *host;
  if ((host = gethostbyname(full_loc_hname)) == (struct hostent *)0) {
#ifdef VERBOSE
    Print_Error("sock::recv_auth", "%s: unknown host", full_loc_hname);
#endif
    return NULL;
  }
  memcpy((char *) &loc_sockaddr->sin_addr, host->h_addr, sizeof(loc_sockaddr->sin_addr));
#endif /* BROKEN_STREAMS_SOCKETS */ 
  
  int sockaddr_in_len = sizeof(struct sockaddr_in);
  if (getsockname(sock, (struct sockaddr *) loc_sockaddr, &sockaddr_in_len) < 0) {
#ifdef VERBOSE
    Print_Error("sock::recv_auth", "%s while getting sock name", strerror(errno));
#endif
    ::close(sock);
    return NULL;
  }
  
#ifdef DEBUG_SOCK
  Print_Error("sock::recv_auth", "loc_sockaddr.sin_addr is %s", inet_ntoa(loc_sockaddr->sin_addr));
  Print_Error("sock::recv_auth", "loc_sockaddr.sin_port is %d", ntohs(loc_sockaddr->sin_port));
#endif

  if(is_kerb5()) {
    
#ifdef KRB5
    krb5_keytab keytab;
    if (retval = krb5_kt_resolve(context, keytab_name, &keytab)) {
      com_err("sock::recv_auth", retval,"while resolving keytab file %s", keytab_name);
      return NULL;
    }
    
    if (!service) {
#ifdef DEBUG_SOCK
      Print_Error("sock::recv_auth","Proceeding with out service name");
#endif
    }
    krb5_principal server;
    if (retval = krb5_sname_to_principal(context, NULL, service, 
					 KRB5_NT_SRV_HST, &server)) {
      com_err("sock::recv_auth", retval,
	      "while generating service name (%s)", service);
      return NULL;
    }
    
    krb5_ticket * ticket;
    if (retval = krb5_recvauth(context, &auth_context, (krb5_pointer)&sock,
			       KRB5_SOCK_VERSION, server, 
			       0,	/* no flags */
			       keytab,	/* default keytab is NULL */
			       &ticket)) {
      com_err("sock::recv_auth", retval, "recvauth failed");
      return NULL;
    }
    
    char *cname;
    if (retval = krb5_unparse_name(context, ticket->enc_part2->client, &cname)){
      com_err("sock::recv_auth", retval, "unparse failed");
#ifdef DEBUG_SOCK
      Print_Error("sock::recv_auth", "You are <unparse error>");
#endif
    } else {
#ifdef DEBUG_SOCK
      Print_Error("sock::recv_auth", "You are %s", cname);
#endif
    }
    
    krb5_address rmt_addr;
    krb5_address loc_addr;
    
    rmt_addr.addrtype = ADDRTYPE_IPPORT;
    rmt_addr.length = sizeof(rmt_sockaddr->sin_port);
    rmt_addr.contents = (krb5_octet *) &rmt_sockaddr->sin_port;
    
    loc_addr.addrtype = ADDRTYPE_IPPORT;
    loc_addr.length = sizeof(loc_sockaddr->sin_port);
    loc_addr.contents = (krb5_octet *) &loc_sockaddr->sin_port;
    
    if ((retval = krb5_auth_con_setports(context, auth_context,
					 &loc_addr, &rmt_addr))) {
      com_err("sock::recv_auth", retval, "while setting remote port\n");
      return NULL;
    }
    
    rmt_addr.addrtype = ADDRTYPE_INET;
    rmt_addr.length = sizeof(rmt_sockaddr->sin_addr);
    rmt_addr.contents = (krb5_octet *) &rmt_sockaddr->sin_addr;
    
    loc_addr.addrtype = ADDRTYPE_INET;
    loc_addr.length = sizeof(loc_sockaddr->sin_addr);
    loc_addr.contents = (krb5_octet *) &loc_sockaddr->sin_addr;
    
    if ((retval = krb5_auth_con_setaddrs(context, auth_context,
					 &loc_addr, &rmt_addr))) {
      com_err("sock::recv_auth", retval, "while setting remote addr\n");
      return NULL;
    }
    
    return cname;
#endif
  }
  else {
#ifdef DEBUG_SOCK
    Print_Error("sock::recv_auth", "Not using Kerberos");
#endif

    return "NO_KRB5_";
  }
}


/**********
Function: sock::read_wait
Purpose:  Select on socket with timeout.  Calling will wait on new socket
          data or timeout on xmit_time, 0 xmit_time disables.
***********/

int sock::read_wait(void)
{
  int n;
  if (xmit_time == 0)
    return 1;
  
  do {
    fd_set readfds;
    struct timeval tv;
    FD_ZERO(&readfds);
    FD_SET(sock, &readfds);
    
    tv.tv_sec = xmit_time;
    tv.tv_usec = 0;

    n = select((sock+1), &readfds, (fd_set *) NULL, (fd_set *) NULL, &tv);

    if (n > 0) 
      return 1;
#ifdef DEBUG_SOCK
    Print_Error("sock::read_wait", "select: Interrupted system call");
#endif
  } while (n < 0 && errno == EINTR);

  if (n == 0) {
    errno = ETIMEDOUT;
#ifdef VERBOSE
    Print_Error("sock::read_wait", "Read timeout after %d seconds.", xmit_time);
#endif
  }
#ifdef VERBOSE
  Print_Error("sock::read_wait", "%s while waiting on socket", strerror(errno));
#endif
  return 0;
}


/**********
Function: sock::read
Purpose:  Reads the data from the socket unencryped and source verified then,
          puts it in the buffer.
          Returns 0 on error, 1 otherwise
**********/
int sock::read(char *buf, size_t buflen)
{
  bzero((void *) buf, buflen);

  if (!read_wait()) {
    return 0;
  }
  
  if (kerberized){
#ifdef KRB5
    krb5_data packet;
    krb5_data message;
    short xmitlen;
    
    if ((retval = krb5_net_read(context, sock, (char *)&xmitlen,
				sizeof(xmitlen))) <= 0) {
      if (retval == 0)
	errno = ECONNABORTED;
      com_err("sock::read", errno, 
	      "while reading data from server");
      return 0;
    }
    packet.length = ntohs(xmitlen);
    if (!(packet.data = (char *)malloc(packet.length + 1))) {
      com_err("sock::read", ENOMEM,
	      "while allocating buffer to read from server");
      return 0;
    }
    if ((retval = krb5_net_read(context, sock, (char *)packet.data,
				packet.length)) <= 0) {
      if (retval == 0)
	errno = ECONNABORTED;
      com_err("sock::read", errno, 
	      "while reading data from server");
      return 0;
    }
    
#ifdef DEBUG_SOCK
    Print_Error("sock::read", "Received %d bytes", packet.length);
#endif
    
    if ((retval = krb5_rd_priv(context, auth_context, &packet,
			     &message, NULL))) {
      com_err("sock::read", retval, 
	      "while verifying PRIV message");
      return 0;
    }
    
    free((void *)packet.data);
    
    strncpy(buf, message.data, buflen);
    return message.length;
#endif
  }
  else {
    short xmitlen;
    int ret_val;
    char *packet;
    int packet_len;

    if ((ret_val = ::read(sock, (char *)&xmitlen, sizeof(xmitlen))) <= 0) {
      if (ret_val == 0)
	errno = ECONNABORTED;
#ifdef VERBOSE
      Print_Error("sock::read", "%s: while reading data from server", strerror(errno));
#endif
      return 0;
    }
    packet_len = ntohs(xmitlen);
    if (!(packet = (char *)malloc(packet_len + 1))) {
#ifdef VERBOSE
      Print_Error("sock::read","%s while allocating buffer to read from server",strerror(ENOMEM));
#endif
      return 0;
    }
    bzero((void *) packet, packet_len + 1);
    if ((ret_val = ::read(sock, (char *)packet, packet_len)) <= 0) {
      if (ret_val == 0)
	errno = ECONNABORTED;
#ifdef VERBOSE
      Print_Error("sock::read", "%s while reading data from server", strerror(errno));
#endif
      return 0;
    }
    
#ifdef DEBUG_SOCK
    Print_Error("sock::read", "Received %d bytes", packet_len);
#endif
    
    strncpy(buf, packet, buflen);
  
    free((void *)packet);
    return packet_len;
  }
}


/**********
Function: sock::write
Purpose:  Writes message to the socket encryped and source verified then,
          Returns 1 on error, 0 otherwise
**********/
int sock::write(char *buf)
{
  if (kerberized) {

#ifdef KRB5
    krb5_data packet;
    krb5_data inbuf;
    short xmitlen;
    
    inbuf.data = buf;
    inbuf.length = strlen(buf);
    
    if ((retval = krb5_mk_priv(context, auth_context, &inbuf,
			       &packet, NULL))) {
      com_err("sock::write", retval, 
	      "while making KRB_PRIV message");
      return 0;
    }
    
    xmitlen = htons(packet.length);
    if ((retval = krb5_net_write(context, sock, (char *)&xmitlen,
				 sizeof(xmitlen))) < 0) {
      com_err("sock::write", errno, 
	      "while writing len to client");
      return 0;
    }
    if ((retval = krb5_net_write(context, sock, (char *)packet.data,
				 packet.length)) < 0) {
      com_err("sock::write", errno,
	      "while writing data to client");
      return 0;
    }
    
    krb5_xfree(packet.data);
    
    return packet.length;
#endif
  }
  else {
    short xmitlen;
    int buflen = strlen(buf);
    int ret_val;
    
    xmitlen = htons(buflen);
    if ((ret_val = ::write(sock, (char *)&xmitlen, sizeof(xmitlen))) < 0) {
#ifdef VERBOSE
      Print_Error("sock::write", "%s while writing len to client", strerror(errno));
#endif
      return 0;
    }
    if ((ret_val = ::write(sock, (char *)buf, buflen)) < 0) {
#ifdef VERBOSE
      Print_Error("sock::write", "%s while writing data to client", strerror(errno));
#endif
      return 0;
    }
    
    return buflen;
  }
}

