/******************************************************************
sock_cli.C
HummingBird Project - Deborah A. Frincke, PhD.
Sprint 1998
Coded by: Jamie Marconi
Description: This class is used for client connections.  It
is also used to hand off server connection to a socket after a connection
has been authenicated.
*******************************************************************/
/*
 Copyright (C) 1997-1998 The HummingBird Project - hummer@cs.uidaho.edu

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "sock_cli.h"

/*******
Function: sock_cli::sock_cli
Purpose:  Basic constructor. A few variable inits.
*******/

sock_cli::sock_cli()
{
  keytab_name = NULL;
  principal_name = NULL;
}

/**********
Function: sock_cli::sock_cli
Purpose:  Constructor that takes a new file descriptor. Used when passing
          off connections from a server socket.
**********/

sock_cli::sock_cli(int asock)
{
  keytab_name = NULL;
  principal_name = NULL;

  sock = asock;
}

/**********
Function: sock_cli::sock_cli
Purpose:  Constructor that takes a new file descriptor, and service name. Used when passing
          off connections from a server socket.
**********/

sock_cli::sock_cli(int asock, char * pservice)
{
  keytab_name = NULL;
  principal_name = NULL;
  service = strdup(pservice);
  
  sock = asock;
}

/**********
Function: sock_cli::sock_cli
Purpose:  Constructor that takes the hostname and the port of the machine 
       to connect to. Calls Create and Connect.
**********/

sock_cli::sock_cli(char *remote_host, int port)
{
  keytab_name = NULL;
  principal_name = NULL;
  
  create();
  connect(remote_host, port);
}


/**********
Function: sock_cli::sock_cli
Purpose:  Constructor that takes the hostname, and the service of the machine 
       to connect to. Calls Create and Connect.
**********/

sock_cli::sock_cli(char *remote_host, char *pservice)
{
  keytab_name = NULL;
  principal_name = NULL;
  
  create();
  connect(remote_host, pservice);
}

/**********
Function: sock_cli::sock_cli
Purpose:  Constructor that takes the hostname, the port, and the service of the machine 
       to connect to. Calls Create and Connect.
**********/
sock_cli::sock_cli(char *remote_host, int port, char *pservice)
{
  keytab_name = NULL;
  principal_name = NULL;
  
  create();
  connect(remote_host, port, pservice);
}

/*******
Function: sock_cli::~sock_cli
Purpose:  Basic destructor
********/

sock_cli::~sock_cli()
{
  if (service)
    free((void*) service);

#ifdef KRB5
  if (auth_context) 
    krb5_auth_con_free(context, auth_context);

  auth_context = NULL;
#endif
}

/*********
Fucntion: sock_cli::set_names
Purpose:  fill in the client keytab, and principal names
*********/

void sock_cli::set_names(char *keytab, char *principal)
{
  keytab_name = strdup(keytab);
  principal_name = strdup(principal);
}

/**********
Function: sock_cli::connect
Purpose:  Connects to the hostname and port specified with service.
          Fills in all host information, local and remote.
          Returns 0 on error, 1 otherwise
**********/

int sock_cli::connect(char *remote_host, int port)
{
  if (!service) {
#ifdef DEBUG_SOCK
    Print_Error("sock_cli::connect", "No Service Name Specified");
#endif
  } 

  // Make sure we got a remote hostname
  if (!remote_host) {
#ifdef VERBOSE
    Print_Error("sock_cli::connect", "Invalid remote hostname");
#endif
    return 0;
  }

  // Look up fully qualified remote hostname
  struct hostent *host;
  if ((host = gethostbyname(remote_host)) == (struct hostent *) 0) {
#ifdef VERBOSE
    Print_Error("sock_cli::connect", "%s: unknown host", remote_host);
#endif
    return 0;
  }

  strncpy(full_rmt_hname, host->h_name, sizeof(full_rmt_hname)-1);
  full_rmt_hname[sizeof(full_rmt_hname)-1] = '\0';

  for (char *cp = full_rmt_hname; *cp; cp++)
    if (isupper(*cp))
      *cp = tolower(*cp);

 // Fill out a sockaddr_in for the remote hostname
  memset((char *) rmt_sockaddr, 0, sin_len);
  memcpy((char *) &rmt_sockaddr->sin_addr, host->h_addr, sizeof(rmt_sockaddr->sin_addr));

  // and the port and type
  rmt_sockaddr->sin_family = AF_INET;
  rmt_sockaddr->sin_port = htons(port);
  
#ifdef DEBUG_SOCK
  Print_Error("sock_cli::connect", "rmt_sockaddr.sin_addr is %s", inet_ntoa(rmt_sockaddr->sin_addr));
  Print_Error("sock_cli::connect", "rmt_sockaddr.sin_port is %d", ntohs(rmt_sockaddr->sin_port));
#endif

  // Connect to Remote Host
  
  if (::connect(sock, (struct sockaddr *)rmt_sockaddr, sin_len) < 0) {
#ifdef VERBOSE
    Print_Error("sock_cli::connect","%s: while connecting to %s", strerror(errno), full_rmt_hname);
#endif
    ::close(sock);
    return 0;
  }

  return 1;
}


/**********
Function: sock_cli::connect
Purpose:  Connects to the hostname  specified with pservice.
          Fills in all host information, local and remote.
          Returns 0 on error, 1 otherwise
**********/

int sock_cli::connect(char *remote_host, char* pservice)
{
  struct servent *serv;

  if (!pservice) {
#ifdef VERBOSE
    Print_Error("sock_cli::connect","empty service name without port");
#endif
    return 0;
  } 

  if ((serv = getservbyname(pservice, "tcp")) == NULL) {
#ifdef VERBOSE
    Print_Error("sock_cli::connect","service unknown: %s/tcp", pservice);
#endif
    return 0;
  }

  return connect(remote_host, ntohs(serv->s_port), pservice);
}

/**********
Function: sock_cli::connect
Purpose:  Connects to the hostname  specified with pservice and port.
          Fills in all host information, local and remote.
          Returns 0 on error, 1 otherwise
**********/

int sock_cli::connect(char *remote_host, int port, char *pservice)
{
  service = strdup(pservice);
  
  return connect(remote_host, port);
}

